#!/bin/sh
#
# Copyright (C) 2022  Braiins Systems s.r.o.
#
# This file is part of Braiins Open-Source Initiative (BOSI).
#
# BOSI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Please, keep in mind that we may also license BOSI or any part thereof
# under a proprietary license. For more information on the terms and conditions
# of such proprietary license or if you have any other questions, please
# contact us at opensource@braiins.com.

set -e -o pipefail
set -x

ROOT="$PWD"
BASEDIR=$(dirname "$0")
. "$BASEDIR/common.sh"

BOS_ENV_FILE="$BASEDIR/$BOS_ENV_FILE"
BOS_UBOOT_FILE="$BASEDIR/$BOS_UBOOT_FILE"
BOS_DTB_FILE="$BASEDIR/$BOS_DTB_FILE"
BOS_KERNEL_FILE="$BASEDIR/$BOS_KERNEL_FILE"
BOS_ROOTFS_FILE="$BASEDIR/$BOS_ROOTFS_FILE"
BOS_TOOLS_NAME="$BASEDIR/$BOS_TOOLS_NAME"
POST_INSTALL_DIR="$BASEDIR/$POST_INSTALL_DIR"
GZIP="$BASEDIR/gzip"
CPIO="$BASEDIR/cpio"

PROC_MTD_SHA256="6a205898fa643c0608182e7e42d3d295ead2d49354d4885e26a5364118d3126a"
HW_MODEL_NAME="TI AM335x BeagleBone"
CPU_ARCH_NEEDED="armv7l"
TOOLS_NEEDED="umount flash_eraseall nandwrite nanddump sha256sum grep uname nohup cp tar"

# In `/tmp` directory the installation is already unpack, so only space for
# stock `config` partition and `nand_env` is needed. Plus some margin (2MB).
# It must comfortably fit 2x content of stock `config` partition and `nand_env`
# Units of `REQUIRED_FREE_SPACE_TMP` are in KB same as `df` output
REQUIRED_FREE_SPACE_TMP=2000
SIZE_OF_CONFIG=$(df | grep '/config' | awk '{print $4}')
REQUIRED_FREE_SPACE_TMP=$((REQUIRED_FREE_SPACE_TMP + 2 * $SIZE_OF_CONFIG))

NVDATA_PARTITION_NUMBER=11
NVDATA_PARTITION="/dev/mtd$NVDATA_PARTITION_NUMBER"

STOCK_UBOOT_PARTITON_NUMBER=4
STOCK_UBOOT_PARTITION="/dev/mtd$STOCK_UBOOT_PARTITON_NUMBER"

STOCK_UBOOT_ENV_PARTITON_NUMBER=5
STOCK_UBOOT_ENV_PARTITION="/dev/mtd$STOCK_UBOOT_ENV_PARTITON_NUMBER"

# Temp directory to mount the rootfs
TMP_BOS_DIR="/tmp/bos_image"
# Define bos_platform file
BOS_PLATFORM_PATH="etc/bos_platform"
# Define bosminer file
BOSMINER_PATH="usr/bin/bosminer"

# Calculate local offset from global partition offset and global desired offset
local_offset() {
    local mtd_partition_global_offset=$1
    local global_offset=$2
    printf "0x%X" $(($global_offset - $mtd_partition_global_offset))
}

# Get local offsets
if [ ! -f "/proc/mtd" ]; then
    echo "Error: Missing /proc/mtd" >&2
    exit 1
fi

# absolute offset of `nvdata` partition in NAND
GLOBAL_NVDATA_OFFSET=0x02000000
# On which offset write installation files to device $NVDATA_PARTITION
# to have it on global offsets $GLOBAL_BOS_*_OFFSET
LOCAL_BOS_UBOOT_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_BOS_UBOOT_OFFSET)
LOCAL_BOS_DTB_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_BOS_DTB_OFFSET)
LOCAL_BOS_KERNEL_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_BOS_KERNEL_OFFSET)
LOCAL_BOS_ROOTFS_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_BOS_ROOTFS_OFFSET)
LOCAL_ENV_TAR_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_ENV_TAR_OFFSET)
LOCAL_ENV_TAR_LENGTH_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_ENV_TAR_LENGTH_OFFSET)
LOCAL_BACKUP_NAND_ENV_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_NAND_ENV_OFFSET)
LOCAL_RECOVERY_FLAGS_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RW_FLAGS_OFFSET)
LOCAL_RO_POST_INSTALL_TAR_LENGTH_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_POST_INSTALL_TAR_LENGTH_OFFSET)
LOCAL_RO_POST_INSTALL_TAR_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_POST_INSTALL_TAR_OFFSET)

# Add `trap_handler` to remove temp install directory
trap_handler() {
    rc="$?"

    # Remove the temp directory only if ends with error
    # In case of end with success, the temp directory needs to be kept
    # for second stage of installation
    if [ "$rc" -eq 0 ]; then
        exit 0
    fi
    echo "Removing temp install directory"
    rm -rf "$(realpath $BASEDIR)"

    exit $rc
}

trap "trap_handler" EXIT

# test command presence and grep command help output for all given arguments
test_command() {
    local cmd=$1
    shift
    if ! [ -x "$(command -v $cmd)" ]; then
        return 1
    fi
    while ! [ -z "$1" ]; do
        "$cmd" -h 2>&1 | grep -q "$1" || return 1
        shift
    done
}

stop_and_kill_service() {
    local service_name=$1
    shift
    /etc/init.d/$service_name stop &
    sleep 1
    while ! [ -z "$1" ]; do
        local process_name=$1
        killall -9 $process_name 2>/dev/null || true
        shift
    done
}

# Get file size in decimal number if second parameter is 10
# otherwise return files size in hexadecimal number
file_size() {
    number_base=$2
    decimal_file_size="$(wc -c < "$1")"
    if [ "$number_base" = "10" ]; then
        echo "$decimal_file_size"
        return 0
    fi
    printf "0x%x" "$decimal_file_size"
}

# check we're running on the right platform and all prerequisities are met
preinstall_check() {
    # check we are not running BOS already
    if [ -f /etc/bos_version ]; then
        echo "Error: BOS is already installed" >&2
        return 1
    fi

    # check required files are available
    for i in $BOS_ROOTFS_IMAGE $BOS_DTB_FILE $BOS_ENV_FILE $BOS_UBOOT_IMAGE $BOS_KERNEL_IMAGE $BOS_TOOLS_NAME; do
        if [ ! -f $i ]; then
            echo "Error: file $i required for installation is missing." >&2
            return 1
        fi
    done

    local tmp_free_space=$(df | grep 'tmpfs.*/tmp' | awk '{print $4}')
    if [ $REQUIRED_FREE_SPACE_TMP -gt "$tmp_free_space" ];  then
        echo "Error: Insufficient space on /tmp." >&2
        exit 1
    fi

    # check required tools are available
    for i in $TOOLS_NEEDED; do
        if ! test_command $i; then
            echo "Error: $i is not installed." >&2
            return 1
        fi
    done

    # check available nandwrite parameters
    if ! test_command nandwrite "\-m" "\-s addr" "\-p" \
        "nandwrite \[OPTION\] MTD_DEVICE \[INPUTFILE\|\-\]"; then
        echo "Error: incompatible nandwrite" >&2
        return 1
    fi

    # check available nanddump parameters
    if ! test_command nanddump "\-\-bb" "skipbad" "\-s addr" "\-l" \
        "nanddump \[OPTIONS\] MTD-device"; then
        echo "Error: incompatible nanddump" >&2
        return 1
    fi

    # check NAND layout
    if ! sha256sum /proc/mtd | grep -q "$PROC_MTD_SHA256"; then
        echo "Error: Unsupported NAND layout" >&2
        cat /proc/mtd >&2
        sha256sum /proc/mtd
        return 2
    fi

    # Check architecture
    if ! uname -a | grep -q "$CPU_ARCH_NEEDED"; then
        echo "Error: Unsupported architecture (must be $CPU_ARCH_NEEDED)." >&2
        return 2
    fi

    # check that the control board is Antminer
    if ! grep -q Antminer "/usr/bin/compile_time"; then
        echo "Error: This install script is only for Antminer." >&2
        exit 2
    fi

    # check board model name
    if ! cat /proc/device-tree/model | grep -q "$HW_MODEL_NAME"; then
        echo "Unsupported HW MODEL (must be ${HW_MODEL_NAME})" >&2
        return 2
    fi

    # Check U-Boot version
    if ! nanddump -l 65 "$STOCK_UBOOT_PARTITION" | grep -a -q "U-Boot 2013.04-g2060661"; then
        echo "Unexpected U-Boot version" >&2
        # List all strings from the whole partition with the prefix 'U-Boot '
        # to try to find the current version of U-Boot
        nanddump "$STOCK_UBOOT_PARTITION" | grep -a "U-Boot " >&2
        return 2
    fi

    # Check miner
    chmod +x "$GZIP" "$CPIO"
    # Extract bosminer and platform from rootfs.img, directory will be created
    platform=$(tail -c+65 < "$BOS_ROOTFS_FILE" | $GZIP -d | $CPIO --to-stdout -i ${BOS_PLATFORM_PATH})
    tail -c+65 < "$BOS_ROOTFS_FILE" | $GZIP -d | $CPIO --directory="$TMP_BOS_DIR" -id ${BOSMINER_PATH}

    # Run bosminer with model detect
    ${TMP_BOS_DIR}/${BOSMINER_PATH} --platform "$platform" model --installation-compatibility-check || bosminer_check=$?
    # Remove temp directory
    rm -rf "$TMP_BOS_DIR"
    # Check bosminer exit code if it is not 0, then miner is not supported
    if [ "$bosminer_check" -ne 0 ]; then
        echo "Error: Miner is not supported by BOS." >&2
        exit 3
    fi
}

write_data() {
    # Stop daemons
    stop_and_kill_service S50lighttpd lighttpd
    stop_and_kill_service S71monitorcg S71monitorcg
    stop_and_kill_service S70cgminer cgminer bmminer
    stop_and_kill_service S52miner_act S52miner_act
    stop_and_kill_service S65monitor-ipsig monitor-ipsig
    stop_and_kill_service S66monitor-recobtn monitor-recobtn
    stop_and_kill_service S01syslogd syslogd
    stop_and_kill_service S41ntp ntpd
    stop_and_kill_service S50avahi-daemon avahi-daemon

    # unmount boot partition used for logs on stock
    local bootpart_mtdblock="/dev/mtdblock$NVDATA_PARTITION_NUMBER"
    if grep -q "$bootpart_mtdblock" /proc/mounts; then
        if ! umount $bootpart_mtdblock; then
            echo "Error: unmount $bootpart_mtdblock" >&2
            return 1
        fi
    fi

    # 1. Flash BOS image to $NVDATA_PARTITION
    # Use this partition because it contains only non-volatile data mainly logs
    flash_eraseall $NVDATA_PARTITION || return 1
    nandwrite -p -s $LOCAL_BOS_UBOOT_OFFSET $NVDATA_PARTITION $BOS_UBOOT_FILE || return 1
    nandwrite -p -s $LOCAL_BOS_DTB_OFFSET $NVDATA_PARTITION $BOS_DTB_FILE || return 1
    nandwrite -p -s $LOCAL_BOS_KERNEL_OFFSET $NVDATA_PARTITION $BOS_KERNEL_FILE || return 1
    nandwrite -p -s $LOCAL_BOS_ROOTFS_OFFSET $NVDATA_PARTITION $BOS_ROOTFS_FILE || return 1

    # 2 Backup nand_env script and /config
    nanddump "$STOCK_UBOOT_ENV_PARTITION" -f "$NAND_ENV_STOCK_FILE" || return 1
    nandwrite -p -s $LOCAL_BACKUP_NAND_ENV_OFFSET $NVDATA_PARTITION "$NAND_ENV_STOCK_FILE" || return 1
    # Create only tar file, because stock do not have gzip
    env_path=/tmp/env
    mkdir -p "$env_path" || return 1
    cp -aR /config "$env_path/$BOS_ENV_STOCK_CONFIG" || return 1
    # Create factory configuration and add to env tar
    $BOS_TOOLS_NAME create-factory-conf -o "$env_path/bos_factory.json"
    tar -cvf $ENV_TAR -C "$env_path" . || return 1
    printf "0x%.8x\0" "$(file_size $ENV_TAR)" | nandwrite -p -s $LOCAL_ENV_TAR_LENGTH_OFFSET $NVDATA_PARTITION
    nandwrite -p -s $LOCAL_ENV_TAR_OFFSET $NVDATA_PARTITION $ENV_TAR || return 1

    # Create post_install directory
    mkdir -p "$POST_INSTALL_DIR"
    # Create stock configuration and put it in post_install directory
    $BOS_TOOLS_NAME create-stock-conf -o "$POST_INSTALL_DIR/stock_config.json"
    # Copy run.sh to post_install directory
    cp "$BASEDIR/run.sh" "$POST_INSTALL_DIR"

    # Check if post_install have ipk packages
    packages_dir="$POST_INSTALL_DIR/packages/"
    if [ -d "$packages_dir" ]; then
        # Remove all non `.ipk` files
        cd "$packages_dir"
        # Subcommand can't be in quote to delete each file
        rm -rf $(ls | grep -v '\.ipk')
        cd "$ROOT"
    fi

    # Store post-install dir in nand
    tar -cvf $POST_INSTALL_TAR -C "$BASEDIR" "$(basename "$POST_INSTALL_DIR")"

    decimal_post_install_size=$(file_size "$POST_INSTALL_TAR" 10)
    if [ "$decimal_post_install_size" -gt "$POST_INSTALL_TAR_MAX_DECIMAL_SIZE" ];  then
        echo "Error: $POST_INSTALL_TAR too big to fit in nand space." >&2
        exit 1
    fi

    printf "0x%.8x\0" "$decimal_post_install_size" | nandwrite -p -s $LOCAL_RO_POST_INSTALL_TAR_LENGTH_OFFSET $NVDATA_PARTITION
    nandwrite -p -s $LOCAL_RO_POST_INSTALL_TAR_OFFSET $NVDATA_PARTITION $POST_INSTALL_TAR

    # Mark state as `INSTALLED` in recovery flags
    printf "$(printf  '\\x%x' $RECOVERY_FLAG_INSTALLED)" | nandwrite -p -s $LOCAL_RECOVERY_FLAGS_OFFSET $NVDATA_PARTITION

    cd "$ROOT"
    touch "$INSTALL_LOG"
    # Change access right so user with the name `miner` can read the log in case of failed installation
    chmod 666 "$INSTALL_LOG"
}

write_env() {
    # This function rewrites stock U-Boot env and run at background with nohup.

    # Wait for first part of script to finish,
    # because killall of dropbear also kills child part of this script
    # if parent is still running.
    sleep 1

    # terminate dropbear
    stop_and_kill_service S50dropbear dropbear

    # write BOS uenv
    flash_eraseall "$STOCK_UBOOT_ENV_PARTITION" || return 1
    nandwrite -p "$STOCK_UBOOT_ENV_PARTITION" "$BOS_ENV_FILE"
    local write_return_code=$?

    echo "Removing temp install directory"
    rm -rf "$(realpath $BASEDIR)"

    if write_return_code; then
        echo "Install successful." >&2
        shutdown -r now
        exit 0
    else
        echo "ERROR: Failed to write UENV!" >&2
        shutdown -r now
        exit 1
    fi
    exit 0
}

# Using hidden `--write-env` parameter will exclude checks, and only write to NAND env.
# This feature is being used to perform forking to background. Writing to env is
# critical part  of installation (it decides whether BOS or stock FW will boot on reboot)
# and we want to be sure it's completed even when SSH session is terminated.
if [ "$1" != "--write-env" ]; then
    preinstall_check || exit 1
    write_data || exit 1
    echo "Creating background task to write to $STOCK_UBOOT_ENV_PARTITION. Log can be found in $INSTALL_LOG."
    $BOS_TOOLS_NAME daemonize /bin/sh -- -c "source $0" "$0" "--write-env" >> "$INSTALL_LOG" 2>&1 0</dev/null

    exit 0
else
    write_env || exit 1
fi
